#include <zephyr/kernel.h>

#include <zephyr/random/random.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/hci.h>
#include <bluetooth/gatt_dm.h>
#include <bluetooth/scan.h>
#include <bluetooth/services/nus.h>
#include <bluetooth/services/nus_client.h>

#if defined(CONFIG_LOG)
#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(main, LOG_LEVEL_DBG);

#define CUSTOM_LOG_DBG(...) LOG_DBG(__VA_ARGS__)
#define CUSTOM_LOG_INF(...) LOG_INF(__VA_ARGS__)
#define CUSTOM_LOG_WRN(...) LOG_WRN(__VA_ARGS__)
#define CUSTOM_LOG_ERR(...) LOG_ERR(__VA_ARGS__)
#else
#define CUSTOM_LOG_DBG(...) printk("DBG: "__VA_ARGS__)
#define CUSTOM_LOG_INF(...) printk("INF: "__VA_ARGS__)
#define CUSTOM_LOG_WRN(...) printk("WRN: "__VA_ARGS__)
#define CUSTOM_LOG_ERR(...) printk("ERR: "__VA_ARGS__)
#endif

#define DEFAULT_SLEEP_TIME_MS           1
#define SLEEP_TIME_BEFORE_DATA_MS       1000
#define SLEEP_TIME_AFTER_DISCONNECT_MS  3000
#define SLEEP_TIME_BETWEEN_DATA_MS      5
#define DATA_RESPONSE_TIMEOUT_MS        500
#define SLEEP_TIME_WHEN_TIMEOUT_MS      10000

#define BUF_SIZE                    256
#define BLE_NUS_MIN_DATA_LEN        10
#define BLE_NUS_MAX_DATA_LEN        240

enum {
    STATE_INIT,
    STATE_START_SCAN,
    STATE_SCANNING,
    STATE_CONNECT,
    STATE_CONNECTING,
    STATE_SET_PHY,
    STATE_WAIT_FOR_PHY,
    STATE_SET_CONN_INTERVAL,
    STATE_WAIT_FOR_INTERVAL,
    STATE_SET_DATALEN,
    STATE_WAIT_FOR_DATALEN,
    STATE_SET_MTU,
    STATE_WAIT_FOR_MTU,
    STATE_START_DISCOVERY_SERVICES,
    STATE_WAIT_FOR_SERVICES,
    STATE_SEND_DATA,
    STATE_SENDING,
    STATE_RESPONSE_WAITING,
    STATE_ERROR,
    STATE_DEAD,
};

/* BLE connection callbacks */
static void connected_cb(struct bt_conn *conn, uint8_t conn_err);
static void disconnected_cb(struct bt_conn *conn, uint8_t reason);
static bool param_update_requested_cb(struct bt_conn *conn, struct bt_le_conn_param *param);
static void param_updated_cb(struct bt_conn *conn, uint16_t interval, uint16_t latency, uint16_t timeout);
static void phy_updated_cb(struct bt_conn *conn, struct bt_conn_le_phy_info *param);
static void data_len_updated_cb(struct bt_conn *conn, struct bt_conn_le_data_len_info *info);

/* NUS client callbacks */
static uint8_t nus_data_received_cb(struct bt_nus_client *nus, const uint8_t *data, uint16_t len);
static void nus_data_sent_cb(struct bt_nus_client *nus, uint8_t err, const uint8_t *const data, uint16_t len);
static void nus_unsubscribed_cb(struct bt_nus_client *nus);

/* GATT exchange callback */
static void exchange_func(struct bt_conn *conn, uint8_t err, struct bt_gatt_exchange_params *params);

/* GATT discovery callbacks */
static void discovery_complete_cb(struct bt_gatt_dm *dm, void *context);
static void discovery_service_not_found_cb(struct bt_conn *conn, void *context);
static void discovery_error_cb(struct bt_conn *conn, int err, void *context);

/* Scan filter match callback */
static void scan_filter_match(struct bt_scan_device_info *device_info, struct bt_scan_filter_match *filter_match, bool connectable);

/* BLE NUS response timeout callback */
static void response_timeout_cb(struct k_timer *timer);

/* BLE buffers */
uint8_t ble_tx_buffer[BUF_SIZE];
size_t ble_tx_len;
uint8_t ble_rx_buffer[BUF_SIZE];
size_t ble_rx_len;

/* Flags for BLE events */
bool peripheral_found;
bool is_connected;
bool phy_adjusted;
bool interval_adjusted;
bool data_len_adjusted;
bool mtu_adjusted;
bool services_discovered;
bool nus_subscribed;
bool ble_data_sent;
bool ble_data_response_received;
bool ble_data_received_valid;
bool ble_data_response_timeout;

int64_t tick_send_time;
int64_t tick_response_time;

/* BLE data reponse timeout timer handler */
static struct k_timer ble_response_timeout_timer;

/* Adjusted (optimized) connection parameters */
struct bt_le_conn_param* new_params = BT_LE_CONN_PARAM(6, 6, 0, 50);

/* PHY connection parameters */
struct bt_conn_le_phy_param param = {
    .options       = BT_CONN_LE_PHY_OPT_CODED_S2,
    .pref_tx_phy   = BT_GAP_LE_PHY_CODED,
    .pref_rx_phy   = BT_GAP_LE_PHY_CODED,
};

/* BLE Scan parameters */
struct bt_le_scan_param scan_param = {
    .type       = BT_HCI_LE_SCAN_ACTIVE,
    .options    = BT_LE_SCAN_OPT_NONE,
    .interval   = BT_GAP_SCAN_FAST_INTERVAL,
    .window     = BT_GAP_SCAN_FAST_WINDOW,
};
struct bt_scan_init_param scan_init = {
    .connect_if_match   = 0,
    .scan_param         = &scan_param,
    .conn_param         = NULL,
};
BT_SCAN_CB_INIT(scan_cb, scan_filter_match, NULL, NULL, NULL);

/* BLE Connection */
BT_CONN_CB_DEFINE(conn_callbacks) = {
    .connected           = connected_cb,
    .disconnected        = disconnected_cb,
    .le_param_req        = param_update_requested_cb,
    .le_param_updated    = param_updated_cb,
    .le_phy_updated      = phy_updated_cb,
    .le_data_len_updated = data_len_updated_cb,
};
struct bt_conn* current_conn;
bt_addr_le_t peer_addr;
bt_addr_le_t expected_addr;

/* BLE NUS client service */
struct bt_nus_client_init_param init = {
    .cb = {
        .received = nus_data_received_cb,
        .sent = nus_data_sent_cb,
        .unsubscribed = nus_unsubscribed_cb,
    }};
struct bt_nus_client nus_client;

/* GATT callbacks */
struct bt_gatt_exchange_params exchange_params = {
    .func = exchange_func,
};
struct bt_gatt_dm_cb discovery_cb = {
    .completed          = discovery_complete_cb,
    .service_not_found  = discovery_service_not_found_cb,
    .error_found        = discovery_error_cb,
};

/* ------------------------ BLE connection callbacks ------------------------ */

static void connected_cb(struct bt_conn *conn, uint8_t conn_err)
{
    int res;
    char addr_str[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr_str, sizeof(addr_str));

    if (conn_err != 0)
    {
        CUSTOM_LOG_ERR("Failed to connect to %s (0x%02X)\n", addr_str, conn_err);
        bt_conn_unref(conn);
        current_conn = NULL;
        return;
    }

    if (current_conn != conn)
    {
        CUSTOM_LOG_WRN("Unexpected conn value: %p, old: %p\n", (void*)conn, (void*)current_conn);
    }

    struct bt_conn_info info;
    res = bt_conn_get_info(conn, &info);
    if (res != 0)
    {
        CUSTOM_LOG_ERR("Failed to get connection info (err %d)\n", res);
        return;
    }
    CUSTOM_LOG_INF("Connected to %s, role %hhu, interval %.1f ms, latency %hu, timeout %hu ms, phy %hhu\n",
                   addr_str,
                   info.role,
                   (double)(info.le.interval * 1.25f),
                   info.le.latency,
                   (info.le.timeout * 10),
                   info.le.phy->tx_phy);
    
    if (current_conn == NULL)
    {
        current_conn = bt_conn_ref(conn);
        if (current_conn == NULL)
        {
            CUSTOM_LOG_ERR("Failed to reference connection\n");
        }
    }
    is_connected = true;
}

static void disconnected_cb(struct bt_conn *conn, uint8_t reason)
{
    char addr_str[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr_str, sizeof(addr_str));

    CUSTOM_LOG_WRN("Disconnected from %s (reason 0x%02X)\n", addr_str, reason);

    if (current_conn != NULL)
    {
        bt_conn_unref(current_conn);
        current_conn = NULL;
    }

    memset(&peer_addr, 0, sizeof(peer_addr));
    peripheral_found = false;
    is_connected = false;
    phy_adjusted = false;
    interval_adjusted = false;
    data_len_adjusted = false;
    mtu_adjusted = false;
    services_discovered = false;
    nus_subscribed = false;
}

static bool param_update_requested_cb(struct bt_conn *conn, struct bt_le_conn_param *param)
{
    CUSTOM_LOG_INF("Connection parameters update requested: interval min %.1f ms, interval max %.1f ms, latency %hu, timeout %hu ms\n",
                   (double)(param->interval_min * 1.25f),
                   (double)(param->interval_max * 1.25f),
                   param->latency,
                   (param->timeout * 10));
    return true;
}

static void param_updated_cb(struct bt_conn *conn, uint16_t interval, uint16_t latency, uint16_t timeout)
{
    CUSTOM_LOG_INF("Connection parameters updated: interval %.1f ms, latency %hu, timeout %hu ms\n",
                   (double)(interval * 1.25f),
                   latency,
                   (timeout * 10));
    interval_adjusted = true;
}

static void phy_updated_cb(struct bt_conn *conn, struct bt_conn_le_phy_info *param)
{
    CUSTOM_LOG_INF("PHY updated: tx_phy = %hhu, rx_phy = %hhu\n", param->tx_phy, param->rx_phy);
    phy_adjusted = true;
}

static void data_len_updated_cb(struct bt_conn *conn, struct bt_conn_le_data_len_info *info)
{
    CUSTOM_LOG_INF("Data length updated: tx_max_len %hu, tx_max_time %hu, rx_max_len %hu, rx_max_time %hu\n",
           info->tx_max_len,
           info->tx_max_time,
           info->rx_max_len,
           info->rx_max_time);
    data_len_adjusted = true;
}

/* -------------------------- NUS client callbacks -------------------------- */

static uint8_t nus_data_received_cb(struct bt_nus_client *nus, const uint8_t *data, uint16_t len)
{
    if ((data == NULL) || (len == 0) || (len > sizeof(ble_rx_buffer)))
    {
        CUSTOM_LOG_ERR("Invalid received data length: %u\n", len);
        return BT_GATT_ITER_STOP;
    }
    tick_response_time = k_uptime_get();

    memcpy(ble_rx_buffer, data, len);
    ble_rx_len = len;
    ble_data_response_received = true;

    return BT_GATT_ITER_CONTINUE;
}

static void nus_data_sent_cb(struct bt_nus_client *nus, uint8_t err, const uint8_t *const data, uint16_t len)
{
    if ((err != 0) || (len != ble_tx_len))
    {
        CUSTOM_LOG_ERR("Failed to send data (err %d)\n", err);
    }
    tick_send_time = k_uptime_get();
    ble_data_sent = true;
}

static void nus_unsubscribed_cb(struct bt_nus_client *nus)
{
    CUSTOM_LOG_WRN("NUS TX notifications disabled\n");
    nus_subscribed = false;
}

/* ------------------------- GATT exchange callback ------------------------- */

static void exchange_func(struct bt_conn *conn, uint8_t err, struct bt_gatt_exchange_params *params)
{
    if (err != 0)
    {
        CUSTOM_LOG_ERR("MTU exchange failed (err %" PRIu8 ")\n", err);
        return;
    }
    CUSTOM_LOG_INF("MTU exchange done\n");
    mtu_adjusted = true;
}

/* ------------------------ GATT discovery callbacks ------------------------ */

static void discovery_complete_cb(struct bt_gatt_dm *dm, void *context)
{
    int res;
    struct bt_nus_client* nus = (struct bt_nus_client *)context;
    if (nus != &nus_client)
    {
        CUSTOM_LOG_ERR("Invalid NUS client context\n");
    }
    res = bt_nus_handles_assign(dm, nus);
    if (res != 0)
    {
        CUSTOM_LOG_ERR("Failed to assign NUS handles (err %d)\n", res);
    }
    else
    {
        res = bt_nus_subscribe_receive(nus);
        if (res != 0)
        {
            CUSTOM_LOG_ERR("Failed to subscribe to NUS receive (err %d)\n", res);
        }
        else
        {
            CUSTOM_LOG_INF("NUS subscribed\n");
            nus_subscribed = true;
            services_discovered = true;
        }
    }
    res = bt_gatt_dm_data_release(dm);
    if (res != 0)
    {
        CUSTOM_LOG_ERR("Failed to release discovery data (err %d)\n", res);
    }

    CUSTOM_LOG_INF("Discovery complete\n");
}

static void discovery_service_not_found_cb(struct bt_conn *conn, void *context)
{
    CUSTOM_LOG_ERR("Discovery failed: Service not found\n");
}

static void discovery_error_cb(struct bt_conn *conn, int err, void *context)
{
    CUSTOM_LOG_ERR("Discovery error: %d\n", err);
}

/* ----------------------- Scan filter match callback ----------------------- */

static void scan_filter_match(struct bt_scan_device_info *device_info, struct bt_scan_filter_match *filter_match, bool connectable)
{
    int res;
    char addr_str[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(device_info->recv_info->addr, addr_str, sizeof(addr_str));

    res = bt_addr_le_cmp(device_info->recv_info->addr, &expected_addr);
    if (res != 0)
    {
        CUSTOM_LOG_WRN("Unexpected device found. Addr: %s, connectable: %s\n", addr_str, (connectable) ? "true" : "false");
        return;
    }
    CUSTOM_LOG_INF("Expected device found. Addr: %s, connectable: %s\n", addr_str, (connectable) ? "true" : "false");

    res = bt_scan_stop();
    if ((res != 0) && (res != -EALREADY))
    {
        CUSTOM_LOG_ERR("Failed to stop scanning (err %d)\n", res);
        return;
    }

    bt_addr_le_copy(&peer_addr, device_info->recv_info->addr);

    peripheral_found = true;
}

/* ------------------- BLE NUS response timeout callback -------------------- */

static void response_timeout_cb(struct k_timer *timer)
{
    ble_data_response_timeout = true;
}

/* ---------------------------- Main entry point ---------------------------- */

int main(void)
{
    int res;

    CUSTOM_LOG_INF("Starting Central...\n");

    int state = STATE_INIT;

    for (;;)
    {
        switch (state)
        {
        case STATE_INIT:
            /* BT stack init */
            res = bt_enable(NULL);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Bluetooth init failed (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            CUSTOM_LOG_INF("BT initialized\n");

            /* BT NUS init */
            res = bt_nus_client_init(&nus_client, &init);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to init NUS client (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            CUSTOM_LOG_INF("BT NUS client initialized\n");

            /* Scan init */
            bt_scan_init(&scan_init);
            bt_scan_cb_register(&scan_cb);
            res = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_UUID, BT_UUID_NUS_SERVICE);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to add scan filter (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            res = bt_scan_filter_enable(BT_SCAN_UUID_FILTER, false);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to enable scan filter (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            CUSTOM_LOG_INF("Scan initialized\n");

            /* Timeout timer init */
            k_timer_init(&ble_response_timeout_timer, response_timeout_cb, NULL);

            /* Hardcoded address to connect to (DE:4D:9B:B9:6C:55) */
            expected_addr.type = BT_ADDR_LE_RANDOM;
            expected_addr.a.val[0] = 0x55;
            expected_addr.a.val[1] = 0x6C;
            expected_addr.a.val[2] = 0xB9;
            expected_addr.a.val[3] = 0x9B;
            expected_addr.a.val[4] = 0x4D;
            expected_addr.a.val[5] = 0xDE;
            char tmp_addr_str[BT_ADDR_LE_STR_LEN];
            bt_addr_le_to_str(&expected_addr, tmp_addr_str, sizeof(tmp_addr_str));
            CUSTOM_LOG_INF("Expected address: %s\n", tmp_addr_str);

            state = STATE_START_SCAN;
            break;

        case STATE_START_SCAN:
            /* Start scanning */
            res = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to start scanning (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            CUSTOM_LOG_INF("Scanning...\n");
            state = STATE_SCANNING;
            break;

        case STATE_SCANNING:
            if (peripheral_found == true)
            {
                state = STATE_CONNECT;
            }
            break;

        case STATE_CONNECT:
            res = bt_conn_le_create(&peer_addr,
                                    BT_CONN_LE_CREATE_CONN,
                                    BT_LE_CONN_PARAM_DEFAULT,
                                    &current_conn);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to create connection (err %d)\n", res);
                state = STATE_START_SCAN;
                break;
            }
            CUSTOM_LOG_INF("Connecting to device...\n");
            state = STATE_CONNECTING;
            break;

        case STATE_CONNECTING:
            if ((is_connected == true) && (current_conn != NULL))
            {
                k_sleep(K_MSEC(SLEEP_TIME_BEFORE_DATA_MS));
                state = STATE_SET_PHY;
            }
            break;

        case STATE_SET_PHY:
            CUSTOM_LOG_INF("Requesting PHY update\n");
            res = bt_conn_le_phy_update(current_conn, &param);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to update PHY (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            state = STATE_WAIT_FOR_PHY;
            break;

        case STATE_WAIT_FOR_PHY:
            if (phy_adjusted == true)
            {
                // after PHY update, datalen will also be chagned to default
                data_len_adjusted = false;
                state = STATE_SET_CONN_INTERVAL;
            }
            break;

        case STATE_SET_CONN_INTERVAL:
            CUSTOM_LOG_INF("Requesting connection parameters update\n");
            res = bt_conn_le_param_update(current_conn, new_params);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to update connection parameters: %d\n", res);
                state = STATE_ERROR;
                break;
            }
            state = STATE_WAIT_FOR_INTERVAL;
            break;

        case STATE_WAIT_FOR_INTERVAL:
            if (interval_adjusted == true)
            {
                state = STATE_SET_DATALEN;
            }
            break;

        case STATE_SET_DATALEN:
            CUSTOM_LOG_INF("Requesting data length update\n");
            res = bt_conn_le_data_len_update(current_conn, BT_LE_DATA_LEN_PARAM_MAX);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to update data length (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            state = STATE_WAIT_FOR_DATALEN;
            break;

        case STATE_WAIT_FOR_DATALEN:
            if (data_len_adjusted == true)
            {
                state = STATE_SET_MTU;
            }
            break;

        case STATE_SET_MTU:
            CUSTOM_LOG_INF("Requesting MTU exchange\n");
            res = bt_gatt_exchange_mtu(current_conn, &exchange_params);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to exchange MTU (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            state = STATE_WAIT_FOR_MTU;
            break;

        case STATE_WAIT_FOR_MTU:
            if (mtu_adjusted == true)
            {
                state = STATE_START_DISCOVERY_SERVICES;
            }
            break;

        case STATE_START_DISCOVERY_SERVICES:
            CUSTOM_LOG_INF("Starting service discovery\n");
            res = bt_gatt_dm_start(current_conn, BT_UUID_NUS_SERVICE, &discovery_cb, &nus_client);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to start discovery (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            state = STATE_WAIT_FOR_SERVICES;
            break;

        case STATE_WAIT_FOR_SERVICES:
            if (services_discovered == true)
            {
                k_sleep(K_MSEC(SLEEP_TIME_BEFORE_DATA_MS));
                state = STATE_SEND_DATA;
            }
            break;

        case STATE_SEND_DATA:
            if ((nus_subscribed == false) || (is_connected == false) || (current_conn == NULL))
            {
                CUSTOM_LOG_WRN("BLE connection lost\n");
                k_sleep(K_MSEC(SLEEP_TIME_AFTER_DISCONNECT_MS));
                state = STATE_START_SCAN;
                break;
            }
            do
            {
                ble_tx_len = (sys_rand32_get() % BLE_NUS_MAX_DATA_LEN);
            }
            while (ble_tx_len == 0 || ble_tx_len < BLE_NUS_MIN_DATA_LEN);
            for (size_t i = 0; i < ble_tx_len; ++i)
            {
                ble_tx_buffer[i] = sys_rand8_get();
            }
            ble_data_sent = false;
            ble_data_response_received = false;
            ble_data_response_timeout = false;
            int res = bt_nus_client_send(&nus_client, ble_tx_buffer, ble_tx_len);
            if (res == -ENOTCONN)
            {
                CUSTOM_LOG_WRN("BLE connection lost\n");
                k_sleep(K_MSEC(SLEEP_TIME_AFTER_DISCONNECT_MS));
                state = STATE_START_SCAN;
                break;
            }
            else if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to send data (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            state = STATE_SENDING;
            break;

        case STATE_SENDING:
            if (ble_data_sent == true)
            {
                k_timer_start(&ble_response_timeout_timer, K_MSEC(DATA_RESPONSE_TIMEOUT_MS), K_NO_WAIT);
                state = STATE_RESPONSE_WAITING;
            }
            else if ((nus_subscribed == false) || (is_connected == false) || (current_conn == NULL))
            {
                CUSTOM_LOG_ERR("BLE connection lost\n");
                k_sleep(K_MSEC(SLEEP_TIME_AFTER_DISCONNECT_MS));
                state = STATE_START_SCAN;
            }
            break;

        case STATE_RESPONSE_WAITING:
            if (ble_data_response_received == true)
            {
                k_timer_stop(&ble_response_timeout_timer);

                if (ble_rx_len != ble_tx_len)
                {
                    CUSTOM_LOG_ERR("Invalid data length received\n");
                    k_sleep(K_MSEC(SLEEP_TIME_WHEN_TIMEOUT_MS));
                    state = STATE_SEND_DATA;
                    break;
                }
                size_t backwards_index = ble_rx_len - 1;
                ble_data_received_valid = true;
                for (size_t i = 0; i < ble_rx_len; ++i)
                {
                    if (ble_rx_buffer[i] != ble_tx_buffer[backwards_index - i])
                    {
                        ble_data_received_valid = false;
                    }
                }
                if (ble_data_received_valid == false)
                {
                    CUSTOM_LOG_ERR("Invalid data received\n");
                    k_sleep(K_MSEC(SLEEP_TIME_WHEN_TIMEOUT_MS));
                    state = STATE_SEND_DATA;
                    break;
                }
                int64_t response_time = tick_response_time - tick_send_time;
                CUSTOM_LOG_DBG("> % 3u bytes in % 3lld ms\n", ble_tx_len, response_time);
                k_sleep(K_MSEC(SLEEP_TIME_BETWEEN_DATA_MS));
                state = STATE_SEND_DATA;
                break;
            }
            else if (ble_data_response_timeout == true)
            {
                k_timer_stop(&ble_response_timeout_timer);
                CUSTOM_LOG_WRN("BLE NUS response timeout\n");
                k_sleep(K_MSEC(SLEEP_TIME_WHEN_TIMEOUT_MS));
                state = STATE_SEND_DATA;
                break;
            }
            else if ((nus_subscribed == false) || (is_connected == false) || (current_conn == NULL))
            {
                k_timer_stop(&ble_response_timeout_timer);
                CUSTOM_LOG_WRN("BLE connection lost\n");
                k_sleep(K_MSEC(SLEEP_TIME_AFTER_DISCONNECT_MS));
                state = STATE_START_SCAN;
                break;
            }
            break;

        case STATE_ERROR:
            CUSTOM_LOG_ERR("Error state, stopping...\n");
            state = STATE_DEAD;
            break;

        case STATE_DEAD:
            break;
        }
        k_sleep(K_MSEC(DEFAULT_SLEEP_TIME_MS));
    }

    return 0;
}
